# Environment Variables
ENV_DEV_MODE = 'DYNOSCALE_DEV_MODE'
ENV_HEROKU_DYNO = 'DYNO'
ENV_DYNOSCALE_URL = 'DYNOSCALE_URL'
# Redis
ENV_REDIS_URL = 'REDIS_URL'
ENV_REDISTOGO_URL = 'REDISTOGO_URL'
ENV_OPENREDIS_URL = 'OPENREDIS_URL'
ENV_REDISGREEN_URL = 'REDISGREEN_URL'
ENV_REDISCLOUD_URL = 'REDISCLOUD_URL'
# Header
HTTP_X_REQUEST_START = 'HTTP_X_REQUEST_START'
X_REQUEST_START = 'X-REQUEST-START'
